# encoding: utf-8
# This configuration will be added by the Rails-Application
# Decorator to add some special features to a document

#module Nukleus::Document::Validations

Nukleus::Document.class_eval do
    def self.validation_states
        [:DOCUMENT_VALIDATION_UNSPECIFIED, :DOCUMENT_VALIDATION_DENIED, :DOCUMENT_VALIDATION_ACCEPTED]
    end

    def self.document_states
        [:DOCUMENT_STATE_IMPORTED, :DOCUMENT_STATE_ASSIGNED, :DOCUMENT_STATE_RETURNED]
    end

    def self.[](state_symbol)
        self.validation_states.index(state_symbol)
    end

    belongs_to  :user, :index => true, :class_name => DeviseAuthentication::User
    field       :validation,    :type => Integer,               :default => self[:DOCUMENT_VALIDATION_UNSPECIFIED]
    field       :is_exported,   :type => NoBrainer::Boolean,    :default => false,                              :index => true
    field       :origin_user_id,:type => String,                :default => nil
    field       :history,       :type => Array,                 :default => []
    field       :state,         :type => Integer,               :default => self.document_states.index(:DOCUMENT_STATE_IMPORTED)

    default_scope { order_by(:created_at => :asc) }

    # Action Cable
    after_create    { DocumentBroadcastJob.perform_later cable_status.to_json }
    before_save     { DocumentBroadcastJob.perform_later cable_status.to_json }
    after_destroy   { DocumentBroadcastJob.perform_later cable_status.to_json }


    # Used by sachbearbeiter in order to see all assigned docs
    def self.all_by_user(user)
        return [] if user.nil?

        # Return all documents that are not exported
        query_hash = {is_exported: false}

        # Return only my documents
        query_hash[:user_id] = user.id

        query_hash[:state] = self.document_states.index(:DOCUMENT_STATE_ASSIGNED)

        Nukleus::Document.where(query_hash)
    end

    def self.all_ids_by_user(user)
        self.all_by_user(user).pluck(:id).map{|h| h[:id]}
    end

    def self.all_by_validation(validation, current_user)
        Nukleus::Document.where(validation: validation, is_exported: false, user_id: current_user.id)
    end

    def self.all_validated(current_user)
        fail "buchhalter only" unless current_user.is? :buchhaltung
        Nukleus::Document.where(:validation.not => self[:DOCUMENT_VALIDATION_UNSPECIFIED], :user_id.in => current_user.buchhaltung_user_ids, is_exported: false)
    end

    def self.unassigned_by_validation(validation, current_user)
        # unassigned means: The document was never assigned to any other person
        fail "buchhalter only" unless current_user.is? :buchhaltung
        Nukleus::Document.where(validation: validation, is_exported: false, :state.not.eq => self.document_states.index(:DOCUMENT_STATE_ASSIGNED))
    end

    # Not-Assigned: Documents owned by Buchhaltung
    def self.unassigned(current_user)
        fail "buchhalter only" unless current_user.is? :buchhaltung
        Nukleus::Document.where(is_exported: false, :state.not.eq => self.document_states.index(:DOCUMENT_STATE_ASSIGNED))
    end

    def self.assigned_by_validation(validation, current_user)
        fail "buchhalter only" unless current_user.is? :buchhaltung
        Nukleus::Document.where(validation: validation, is_exported: false, :user_id.in => current_user.sachbearbeiter_user_ids)
    end

    def self.assigned(current_user)
        fail "buchhalter only" unless current_user.is? :buchhaltung
        Nukleus::Document.where(is_exported: false, :user_id.in => current_user.sachbearbeiter_user_ids)
    end

    def self.status_overview(current_user)
        fail "buchhalter only" unless current_user.is? :buchhaltung
        # "new", "assigned" and "finished" are the sections in the web application.
        {
            new: self.unassigned(current_user).pluck(:id).map{|h| h[:id]},
            assigned: self.assigned(current_user).pluck(:id).map{|h| h[:id]},
            finished: self.all_validated(current_user).pluck(:id).map{|h| h[:id]}
        }
    end

    def self.status_user(user_ids, current_user)
        fail "buchhalter only" unless current_user.is? :buchhaltung
        response_data = {}
        user_ids.each do |user_id|
            response_data[user_id] = Nukleus::Document.where(is_exported: false, user_id: user_id).pluck(:id).map{|h| h[:id]}
        end

        response_data
    end

    def set_validation_state(acting_user, validation, comment)
        add_history(acting_user, comment, "Document wurde validiert von #{acting_user.user_name}: #{validation})")
        self.validation = validation
        self.save()
    end

    def assign_to_user(acting_user, dest_user, comment)
        # Add comment to history
        add_history(acting_user, comment, "Document wurde an eine andere Person verschoben (#{acting_user.user_name} -> #{dest_user.user_name})")

        # Save origin user (first importer).
        self.origin_user_id = self.user_id if self.origin_user_id.blank?

        self.state = self.class.document_states.index(:DOCUMENT_STATE_ASSIGNED)

        # Now assign to new user
        self.user_id = dest_user.id

        self.save()
    end

    def back_to_origin(acting_user, comment)
        fail "origin user was not set" if self.origin_user_id.nil?

        origin_user = DeviseAuthentication::User.where(id: self.origin_user_id).first
        fail "origin user does not exists" if origin_user.nil?

        add_history(acting_user, comment, "Document wurde zurück gegeben (#{self.user.user_name} -> #{origin_user.user_name})")

        self.state = self.class.document_states.index(:DOCUMENT_STATE_RETURNED)

        # Now assign to new user
        self.user_id = self.origin_user_id

        self.save()
    end

    def add_comment(acting_user, comment)
        add_history(acting_user, comment, "User #{acting_user} hat ein Kommentar hinzugefügt" )

        self.save()
    end

    def export(acting_user)
        fail "This role cannot export documents" if !acting_user.is? :buchhaltung

        add_history(acting_user, "", "User #{acting_user.user_name} hat das Dokument exportiert" )

        # TODO: Check success state if available
        success = true
        ExportDoc.delay.export(self,"/export")

        if success
            self.is_exported = true
            success = self.save()
        end

        success
    end

    # Returns an image of the first page
    def preview(width, height)
        # Read in first page of PDF and resize ([0] -> read the first page only)
        Rails.cache.fetch("#{self.id}-#{width}-#{height}-#{self.document_version}", expires_in: 1.year) do
            image = nil
            begin
                tempfile = self.document_file.download
                image = ::Magick::Image.read(tempfile.path).first
                image.write("jpg:#{tempfile.path}")
                image.resize_to_fit!(width, height)
                # Most PDFs have a transparent background, which becomes black when converted to jpg.
                # To override this, we must create a white canvas and composite the PDF onto the convas.
                # First check the image backgrounf color.
                # #FFFFFFFFFFFF0000 is how ImageMagick defines transparent
                if image.background_color != "#FFFFFFFFFFFF0000"
                    # Create a blank canvas
                    canvas = ::Magick::Image.new(image.columns, image.rows) { self.background_color = "#FFF" }
                    # Merge PDF thumbnail onto canvas
                    canvas.composite!(image, ::Magick::CenterGravity, ::Magick::OverCompositeOp)
                    # Save as .jpg. We need to change the file extension here so that the fog gem picks it up and
                    # sets the correct mime type. Otherwise the mime type will be set to PDF, which confuses browsers.
                    canvas.write("jpg:#{tempfile.path}")
                    image = canvas.clone
                    canvas.destroy! unless canvas.nil?
                else
                    # If the imag ebackground color is transparent we don't need to worry about the canvas.
                    image.write("jpg:#{tempfile.path}")
                end
            rescue ::Magick::ImageMagickError => e
                Rails.logger.error "Failed to create PDF thumbnail: #{e.message}"
                raise CarrierWave::ProcessingError, "is not a valid PDF file"
            end
            image.format = 'JPEG'
            blob = image.to_blob
            image.destroy!

            blob
        end
    end

    private
    def add_history( acting_user, comment, info )
        self.history.push({
                            comment: comment,
                            info: info,
                            datetime: Time.now,
                            count: self.history.count,
                            user_id: acting_user.id,
                            username: acting_user.user_name,
                            firstname: acting_user.forename,
                            lastname: acting_user.lastname
        })

    end

    def cable_status
        {
            id: self.id,
            validation: self.validation,
            userId: self.user_id,
            state: self.state,
            is_exported: self.is_exported,
            prevUserId: self.user_id_was
        }
    end
end
