require 'rails_helper'
require "refile/file_double"

module RechnMgmWeb
    RSpec.describe Api::V1::ApiDocumentsController, type: :controller do
        routes { RechnMgmWeb::Engine.routes }

        before(:each) {
            # Cleanup
            NoBrainer.purge!
            NoBrainer::Document::Index::Synchronizer.new([Nukleus::Document, DeviseAuthentication::User]).sync_indexes

            # Create some users
            ## Buchhaltung
            @buchhaltung = DeviseAuthentication::User.new(email: "buchhaltung@test.de", forename: "Buchhalter", lastname: "Eins", password: "123456", password_confirmation: "123456" )
            expect(@buchhaltung.save).to eq(true)
            @buchhaltung.member_ships.first.roles_mask = 2
            @organisation_1 = @buchhaltung.member_ships.where(is_primary: true).first.organisation

            ## Sachbearbeiter
            @sachbearbeiter = DeviseAuthentication::User.new(email: "sachbearbeiter@test.de", forename: "Sachbearbeiter", lastname: "Erster", password: "123456", password_confirmation: "123456" )
            expect(@sachbearbeiter.save).to eq(true)
            @sachbearbeiter.current_membership_id =
            RechnMgmUserArea::MemberShip.create(user_id: @sachbearbeiter.id, organisation_id: @organisation_1.id, is_primary: false, roles_mask: 4).id
            # set organisation_1 as current organisation
            @sachbearbeiter.set_organisation(@organisation_1)

            @sachbearbeiter2 = DeviseAuthentication::User.new(email: "sachbearbeiter2@test.de", forename: "Sachbearbeiter", lastname: "Zweiter", password: "123456", password_confirmation: "123456" )
            expect(@sachbearbeiter2.save).to eq(true)
            RechnMgmUserArea::MemberShip.create(user_id: @sachbearbeiter2.id, organisation_id: @organisation_1.id, is_primary: false, roles_mask: 4)
            @sachbearbeiter2.set_organisation(@organisation_1)

            # TODO: Add Sachbearbeiter + Buchhalter..

            # Simulate import of some new documents
            @new_document = create_document(@buchhaltung, "Test1")
            create_document(@buchhaltung, "Test2")
            create_document(@buchhaltung, "Test3")

            # Documenents are assigned to Sachbearbeiter
            @assigned_document = create_document(@sachbearbeiter, "Test4", nil, @buchhaltung)
            create_document(@sachbearbeiter, "Test5", Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED], @buchhaltung)    # Unspecified (explicitly set)
            @accepted_document = create_document(@sachbearbeiter, "Test7", Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED], @buchhaltung)       # Accepted
            create_document(@sachbearbeiter, "Test8", Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung)         # Denied

            # Documenents are assigned to Sachbearbeiter2
            @assigned_document2 = create_document(@sachbearbeiter2, "Test4", nil, @buchhaltung)
            create_document(@sachbearbeiter2, "Test5", Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED], @buchhaltung)    # Unspecified (explicitly set)
            create_document(@sachbearbeiter2, "Test7", Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED], @buchhaltung)       # Accepted
            create_document(@sachbearbeiter2, "Test8", Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung)         # Denied

            # Document is validated and back to Buchahltung
            create_document(@buchhaltung, "Test9", Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED], @buchhaltung)       # Accepted
            create_document(@buchhaltung, "Test10", Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], @buchhaltung)         # Denied
        }

        describe "Check Buchhaltung: " do
            render_views
            before(:each){
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
            }

            it "finds new documents" do
                # New Documents
                get :index_unassigned, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(5)
                expect(body_data.select{|doc| doc["state"] == Nukleus::Document.document_states.index(:DOCUMENT_STATE_IMPORTED)}.count).to eq(3) # 3 are imported
                expect(body_data.select{|doc| doc["state"] == Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED)}.count).to eq(2) # 2 are returned
            end

            it "finds new documents" do
                # New Documents
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(3)
            end

            it "finds documents assigned to sachbearbeiters (in progress)" do
                get :index_assigned, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(8)
            end

            it "finds documents assigned to sachbearbeiter that are not yet validated (in progress, unvalidated)" do
                get :index_assigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(4)
            end

            it "finds documents assigned to other person but are validated (in progess, accepted, not returned)" do
                get :index_assigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
            end

            it "finds documents assigned to other person but are validated (in progess, denied, not returned)" do
                get :index_assigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
            end

            it "finds documents given back to buchhaltung and validated (finished, denied)" do
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED] } , format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(1)
            end

            it "finds documents given back to buchhaltung but are validated (finished and accepted)" do
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(1)
            end

            it "finds all documents he has to work for (assigned)" do
                get :index, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(0) #TODO: Add documuments for users that are sachbearbeiter _and_ buchhalter
            end

            it "finds all validated documents" do
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
            end

            it "receives a status overview" do
                get :status_overview, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)

                expect(body_data.keys.include? "new").to be(true)
                expect(body_data["new"].count).to eq(5)             # 3 Imported, 2 Returned
                expect(body_data["new"].first.class).to be(String)

                expect(body_data.keys.include? "assigned").to be(true)
                expect(body_data["assigned"].count).to eq(8)
                expect(body_data["assigned"].first.class).to be(String)

                expect(body_data.keys.include? "finished").to be(true)
                expect(body_data["finished"].count).to eq(2)
                expect(body_data["finished"].first.class).to be(String)
            end

            it "receives status of other users" do
                get :status_user, :params => {user_ids: "#{@sachbearbeiter.id},#{@sachbearbeiter2.id}"}, format: :json

                expect(response).to have_http_status(:ok)
                body_data = JSON.parse(response.body)

                expect(body_data.class).to be(Hash)
                expect(body_data.keys.include? @sachbearbeiter.id).to be(true)
                expect(body_data["#{@sachbearbeiter.id}"].count).to be(4)
                expect(body_data.keys.include? @sachbearbeiter2.id).to be(true)
                expect(body_data["#{@sachbearbeiter2.id}"].count).to be(4)
            end

            it "set validation state" do
                expect(@assigned_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state, :params => {id: @assigned_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment")
            end

            it "assigns document to user" do
                expect(@new_document.user_id).to eq(@buchhaltung.id)
                expect(@new_document.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_IMPORTED))

                put :assign_to_user, :params => { id: @new_document, user_id: @sachbearbeiter2, comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@new_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@sachbearbeiter2.id)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment")
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))
            end

        end
        describe "Check Sachbearbeiter: " do
            render_views
            before(:each){
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
            }

            it "sees only documents assigned to myself" do
                get :index, { format: :json }
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.class).to be(Array)
                expect(body_data.count).to eq(4)
            end

            it "set validation state" do
                expect(@assigned_document.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state, :params => {id: @assigned_document, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_DENIED])
                expect(test_doc.history.count).to eq(1)
            end

            xit "cannot set validation state if not assigned to him" do
                expect(@assigned_document2.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])

                put :set_validation_state, :params => {id: @assigned_document2, state: Nukleus::Document[:DOCUMENT_VALIDATION_DENIED], comment: "My Comment" }, format: :json
                expect(response).not_to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document2.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.validation).to be(Nukleus::Document[:DOCUMENT_VALIDATION_UNSPECIFIED])
                expect(test_doc.history.count).to eq(1)
            end

            it "does not receives a status overview" do
                expect{
                    get :status_overview, { format: :json }
                }.to raise_error("buchhalter only")
            end

            it "assigns document to user" do
                expect(@assigned_document.user_id).to eq(@sachbearbeiter.id)
                expect(@assigned_document.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))

                put :assign_to_user, :params => { id: @assigned_document, user_id: @sachbearbeiter2, comment: "My Comment 3" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.user_id).to eq(@sachbearbeiter2.id)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment 3")
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED))
            end

            xit "can read document assigned to him" do
                get :show, :params => {id: @assigned_document.id }, format: :json
                expect(response).to have_http_status(:ok)
            end

            xit "cannot read document not assidned to him" do
                get :show, :params => {id: @assigned_document2.id }, format: :json
                expect(response).not_to have_http_status(:ok)
            end

            xit "cannot read preview document not assidned to him" do
                get :preview, :params => { id: @assigned_document2.id, width: "200", height:"400" },  format: :json
                expect(response).not_to have_http_status(:ok)
            end
        end

        describe "Document assignments" do
            render_views
            it "return document back to buchhaltung" do
                # Verwaltung is doing this call. Mock current user..
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                # Check that this document will not be found
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(1)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to eq(false)

                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :back_to_origin, :params => {:id => @accepted_document.id, comment: "My Comment2" }, format: :json
                expect(response).to have_http_status(:ok)

                # Check that this document will now be found
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                get :index_unassigned, :params => { validation: Nukleus::Document[:DOCUMENT_VALIDATION_ACCEPTED] }, format: :json
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to be(true)

                # Check that comment was added.
                test_doc = Nukleus::Document.find(@accepted_document.id)
                expect(test_doc).not_to be(nil)
                expect(test_doc.history.count).to eq(1)
                expect(test_doc.history.first[:comment]).to eq("My Comment2")
                expect(test_doc.state).to eq(Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED))
            end

            # index_validated == validatedDocuments
            it "return document back to buchhaltung for validatedDocuments" do
                # Verwaltung is doing this call. Mock current user..
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                # Check that this document will not be found
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to eq(false)

                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :back_to_origin, :params => {:id => @accepted_document.id }, format: :json
                expect(response).to have_http_status(:ok)

                # Check that this document will now be found
                allow(controller).to receive(:current_user).and_return(@buchhaltung)
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(3)
                expect(body_data.each.map{|d| d["id"]}.include? @accepted_document.id).to be(true)
            end

            # TODO: Check backToOrigin with not validated documents
            it "return document back to buchhaltung for not validatedDocuments" do
                # Verwaltung is doing this call. Mock current user..
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                # Check that this document will not be found
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @assigned_document.id).to eq(false)

                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)
                put :back_to_origin, :params => {:id => @assigned_document.id }, format: :json
                expect(response).to have_http_status(:ok)

                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                # Check that this document cannot be vound as validated document
                get :index_validated, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.count).to eq(2)
                expect(body_data.each.map{|d| d["id"]}.include? @assigned_document.id).to be(false)

                # Check that the document can be found a "new"
                get :index_unassigned, { format: :json }
                expect(response).to render_template("index")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data.each.map{|d| d["id"]}.include? @assigned_document.id).to be(true)
            end
        end

        describe "Document exports" do
            render_views
            it "get preview" do
                allow(controller).to receive(:current_user).and_return(@buchhaltung)

                document = Nukleus::Document.first
                get :preview, :params =>  {id: document.id, width: 600, height: 800}
                expect(response).to have_http_status(:ok)
            end
        end

        describe "Document" do
            render_views

            it "has a history" do
                # Verwaltung is doing this call. Mock current user..
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)

                expect(@assigned_document.add_comment(@sachbearbeiter, "comment 1")).to be(true)
                expect(@assigned_document.add_comment(@sachbearbeiter, "comment 2")).to be(true)
                expect(@assigned_document.add_comment(@sachbearbeiter, "comment 3")).to be(true)

                get :show, :params => { id: @assigned_document.id },  format: :json
                expect(response).to render_template("show")
                expect(response).to have_http_status(:ok)

                body_data = JSON.parse(response.body)
                expect(body_data["history"].count).to eq(3)
            end

            it "add a comment" do
                allow(controller).to receive(:current_user).and_return(@sachbearbeiter)

                initial_history_count = @assigned_document.history.count

                put :add_comment, :params => { id: @assigned_document, comment: "Test" }, format: :json
                expect(response).to have_http_status(:ok)

                test_doc = Nukleus::Document.find(@assigned_document.id)
                expect(test_doc.history.count).to eq(initial_history_count + 1)
                expect(test_doc.history.last["comment"]).to eq("Test")
            end

            # TODO: Add testing for action cable
            # it "created will be signalled with action cable" do
            #     expect { put :add_comment, :params => { id: @assigned_document, comment: "Test" }, format: :json }.to have_broadcasted_to("document_channel") #.with(text: 'Cool!')
            # end
        end


        private

        def create_document(user, file_name, validation_state = nil, origin_user = nil)
            backend = Refile.backends["cache"]

            test_file = File.open("./spec/data/test.pdf")
            file = backend.upload(test_file)

            params = {
                document_file:              file,
                document_id:                SecureRandom.uuid,
                document_file_filename:     file_name.gsub( /[^a-zA-Z0-9_\.]/, '_'),
                # document_file_size:         file.size,
                # document_file_content_type: type,
                user_id:                    user.id,
                meta_data:                  { importer_type: :TEST }
            }

            if origin_user.present?
                # Document is assigned. Thus
                params[:origin_user_id] = origin_user.id
                if user == origin_user
                    params[:state] = Nukleus::Document.document_states.index(:DOCUMENT_STATE_RETURNED)
                else
                    params[:state] = Nukleus::Document.document_states.index(:DOCUMENT_STATE_ASSIGNED)
                end
            end

            if validation_state.present?
                params[:validation] = validation_state
            end

            new_doc = Nukleus::Document.new(params)
            expect(new_doc.save).to eq(true)

            new_doc
        end
    end
end
